# Copyright 2001 Benjamin Trott. This code cannot be redistributed without
# permission from www.movabletype.org.
#
# $Id: DBM.pm,v 1.40 2001/10/08 11:25:15 btrott Exp $

package MT::ObjectDriver::DBM;
use strict;

use DB_File;
use Fcntl;
use File::Spec;

use MT::Util qw( offset_time_list );
use MT::Serialize;
use MT::ObjectDriver;
use base qw( MT::ObjectDriver );

sub init {
    my $driver = shift;
    $driver->SUPER::init(@_);
    $driver->{serializer} = MT::Serialize->new($driver->cfg->Serializer);
    $driver;
}

sub _db_data {
    File::Spec->catfile($_[0]->cfg->DataSource,
        $_[1]->datasource . ".db");
}

sub _db_index {
    File::Spec->catfile($_[0]->cfg->DataSource, $_[1]->datasource .
        '.' . $_[2] . '.idx');
}

sub _tie_db_file {
    my($file, $type) = @_;
    my $created = -e $file ? 0 : 1;
    my $DB = tie my %db, 'DB_File', $file, O_RDWR|O_CREAT, 0666, $type or
        return;
    if ($created) {
        chmod 0666, $file or return;
    }
    ($DB, \%db);
}

sub _get_ids {
    my $driver = shift;
    my($DB, $db, $class, $terms, $args) = @_;
    my @ids;
    if ($args && $args->{limit}) {    ## Lookup with limit
        @ids = $driver->_get_ids_limit($DB, $db, $class, $terms, $args);
    }
    elsif ($terms) {                ## Lookup using index or ID
        if (ref($terms) eq 'HASH') {
            @ids = %$terms ?
                $driver->_get_ids_from_index($class, $terms, $args) :
                keys %$db;
        } else {
            @ids = $terms;
        }
    }
    else {                          ## Lookup all
        @ids = keys %$db;
    }

    ## Now sort if we need to, by sort column. If limit is provided along
    ## with sort, we have already sorted in _get_ids_limit, so we don't
    ## need to do it again.
    if ((my $col = $args->{sort}) && !$args->{limit}) {
        my $direction = $args->{direction} || 'ascend';
        my $idx_file = _db_index($driver, $class, $col);
        my($DB, $idx) = _tie_db_file($idx_file, $DB_BTREE) or
            return $driver->error("Tie '$idx_file' failed: $!");
        my %sort_val = map { $_ => '' } @ids;
        while (my($val, $ids) = each(%$idx)) {
            my @idx_ids = split /$;/, $ids;
            @sort_val{ @idx_ids } = ($val) x @idx_ids;
        }
        @ids = $direction eq 'ascend' ?
            (sort { $sort_val{$a} cmp $sort_val{$b} } @ids) :
            (sort { $sort_val{$b} cmp $sort_val{$a} } @ids);
        undef $DB;
        untie %$idx;
    }
    @ids;
}

sub _get_ids_from_index {
    my $driver = shift;
    my($class, $terms, $args) = @_;
    my %count;
    for my $col (keys %$terms) {
        my $idx_file = _db_index($driver, $class, $col);
        my($DB, $idx) = _tie_db_file($idx_file, $DB_BTREE) or
            return $driver->error("Tie '$idx_file' failed: $!");
        if (ref($terms->{$col}) eq 'ARRAY') {
            if ($args->{range} && $args->{range}{$col}) {   ## Range lookup
                my($start, $end) = @{ $terms->{$col} };
                my($key, $val) = ($start, 0);
                unless ($DB->seq($key, $val, R_CURSOR)) {
                    my @ids = split /$;/, $val || '';
                    my($st);
                    for ($st = $DB->seq($key, $val, R_NEXT);
                         $st == 0 && (!$end || $key < $end);
                         $st = $DB->seq($key, $val, R_NEXT)) {
                        push @ids, split /$;/, $val || '';
                    }
                    for my $id (@ids) { $count{$id}++ }
                }
            }
        }
        else {                                 ## Standard 'equals' lookup
            my $col_value = $terms->{$col} || '';
            my @ids = split /$;/, $idx->{$col_value} || '';
            for my $id (@ids) { $count{$id}++ }
        }
        undef $DB;
        untie %$idx;
    }
    my @ids;
    my $num_cols = scalar keys %$terms;
    for my $id (keys %count) {
        push @ids, $id if $count{$id} >= $num_cols;
    }
    @ids;
}

sub _get_ids_limit {
    my $driver = shift;
    my($DB, $db, $class, $terms, $args) = @_;
    my $n = $args->{limit};
    my $this_db = $DB;
    my $idx;
    my @ids;
    if (my $col = $args->{sort}) {
        my $idx_file = _db_index($driver, $class, $col);
        ($this_db, $idx) = _tie_db_file($idx_file, $DB_BTREE) or
            return $driver->error("Tie '$idx_file' failed: $!");
    }
    my $dir = $args->{direction} || 'ascend';
    my($c1, $c2) = $dir eq 'ascend' ? (R_FIRST, R_NEXT) :
                                      (R_LAST,  R_PREV);
    my($i, $key, $val, $st) = (0, 0, 0);
    if ($args->{start_val}) {
        ## Advance cursor to start val
        $c1 = $dir eq 'ascend' ? R_NEXT : R_PREV;
        $st = $this_db->seq($args->{start_val}, $val, R_CURSOR);
    }
    for ($st = $this_db->seq($key, $val, $c1);
        $st == 0 && $i < $n;
        $st = $this_db->seq($key, $val, $c2)) {
        my @these_ids = $args->{sort} ? split(/$;/, $val) : $key;
        if ($terms) {
            unless ($args->{sort}) {
                my $rec = ${ $driver->{serializer}->unserialize($val) };
                my $matched = 1;
                for my $col (keys %$terms) {
                    $matched = 0, last
                        unless defined($rec->{$col}) &&
                            $terms->{$col} eq $rec->{$col};
                }
                push(@ids, $key), $i++ if $matched;
            } else {
                for my $id (@these_ids) {
                    my $rec = $db->{$id} or next;
                    $rec = ${ $driver->{serializer}->unserialize($rec) };
                    my $matched = 1;
                    for my $col (keys %$terms) {
                        $matched = 0, last
                            unless defined($rec->{$col}) &&
                                $terms->{$col} eq $rec->{$col};
                    }
                    push(@ids, $id), $i++ if $matched;
                }
            }
        } else {
            push(@ids, @these_ids);
            $i++;
        }
    }
    if ($args->{sort}) {
        undef $this_db;
        untie %$idx;
    }
    @ids;
}

sub load_iter {
    my $driver = shift;
    my($class, $terms, $args) = @_;
    my $db_file = _db_data($driver, $class);
    my($DB, $db) = _tie_db_file($db_file, $DB_BTREE) or
        return $driver->error("Tie '$db_file' failed: $!");
    my @ids = $driver->_get_ids($DB, $db, $class, $terms, $args);
    my $idx = 0;
    sub {
        if ($idx > $#ids) {
            undef $DB;
            untie %$db;
            return;
        }
        my $rec = $db->{ $ids[$idx++] } or return;
        $rec = $driver->{serializer}->unserialize($rec);
        my $obj = $class->new;
        $obj->set_values($$rec);
        $obj;
    };
}

sub load {
    my $driver = shift;
    my($class, $terms, $args) = @_;
    my $db_file = _db_data($driver, $class);
    my($DB, $db) = _tie_db_file($db_file, $DB_BTREE) or
        return $driver->error("Tie '$db_file' failed: $!");
    my @ids = $driver->_get_ids($DB, $db, $class, $terms, $args);
    my @objs;
    for my $id (@ids) {
        my $rec = $db->{$id} or return;
        $rec = $driver->{serializer}->unserialize($rec);
        my $obj = $class->new;
        $obj->set_values($$rec);
        return $obj unless wantarray;
        push @objs, $obj;
    }
    undef $DB;
    untie %$db;
    @objs;
}

sub count {
    my $driver = shift;
    my($class, $terms, $args) = @_;
    my $db_file = _db_data($driver, $class);
    my($DB, $db) = _tie_db_file($db_file, $DB_BTREE) or
        return $driver->error("Tie '$db_file' failed: $!");
    my @ids = $driver->_get_ids($DB, $db, $class, $terms, $args);
    undef $DB;
    untie %$db;
    scalar @ids;
}

sub exists {
    my $driver = shift;
    my($obj) = @_;
    return unless $obj->id;
    my $db_file = _db_data($driver, $obj);
    my($DB, $db) = _tie_db_file($db_file, $DB_BTREE) or
        return $driver->error("Tie '$db_file' failed: $!");
    my $exists = exists $db->{$obj->id};
    undef $DB;
    untie %$db;
    $exists;
}

sub save {
    my $driver = shift;
    my($obj) = @_;
    my $db_file = _db_data($driver, $obj);
    my($DB, $db) = _tie_db_file($db_file, $DB_BTREE) or
        return $driver->error("Tie '$db_file' failed: $!");
    unless ($obj->id || ($obj->id($driver->generate_id($obj)))) {
        return $driver->error("Failed to generate unique ID");
    }
    my $id = $obj->id;
    if ($obj->properties->{audit}) {
        my $blog_id = $obj->blog_id;
        my @ts = offset_time_list(time, $blog_id);
        my $ts = sprintf "%04d%02d%02d%02d%02d%02d",
            $ts[5]+1900, $ts[4]+1, @ts[3,2,1,0];
        $obj->created_on($ts)
            unless CORE::exists($db->{$id}) || $obj->created_on;
        $obj->modified_on($ts);
    }
    ## Grab old values so that we can update indexes on changed columns
    my $old = $db->{$id};
    $old = ${ $driver->{serializer}->unserialize($old) } if $old;
    $db->{$id} = $driver->{serializer}->serialize(\$obj->column_values);
    undef $DB;
    untie %$db;

    my $indexes = $obj->properties->{indexes};
    for my $col (keys %$indexes) {
        my $idx_file = _db_index($driver, $obj, $col);
        my($DB, $idx) = _tie_db_file($idx_file, $DB_BTREE) or
            return $driver->error("Tie '$idx_file' failed: $!");
        my $col_value = $obj->$col();
        $col_value = '' unless defined $col_value;
        my %ids = map { $_ => 1 } split /$;/, $idx->{$col_value} || '';
        next if exists $ids{$id};
        $idx->{$col_value} = join $;, keys %ids, $id;
        if ($old && $old->{$col} ne $col_value) {
            _drop_from_index($idx, $id, $old->{$col});
        }
        undef $DB;
        untie %$idx;
    }
    1;
}

sub remove {
    my $driver = shift;
    my($obj) = @_;
    my $id = $obj->id;
    return unless $id;
    my $indexes = $obj->properties->{indexes};
    for my $col (keys %$indexes) {
        my $idx_file = _db_index($driver, $obj, $col);
        my($DB, $idx) = _tie_db_file($idx_file, $DB_BTREE) or
            return $driver->error("Tie '$idx_file' failed: $!");
        my $col_value = $obj->$col();
        _drop_from_index($idx, $id, $col_value);
        undef $DB;
        untie %$idx;
    }
    my $db_file = _db_data($driver, $obj);
    my($DB, $db) = _tie_db_file($db_file, $DB_BTREE) or
        return $driver->error("Tie '$db_file' failed: $!");
    delete $db->{$obj->id};
    undef $DB;
    untie %$db;
    1;
}

sub remove_all {
    my $driver = shift;
    my($class) = @_;
    my $indexes = $class->properties->{indexes};
    for my $col (keys %$indexes) {
        my $idx_file = _db_index($driver, $class, $col);
        next unless -e $idx_file;
        unlink $idx_file or
            return $driver->error("Unlink of '$idx_file' failed: $!");
    }
    my $db_file = _db_data($driver, $class);
    if (-e $db_file) {
        unlink $db_file or
            return $driver->error("Unlink of '$db_file' failed: $!");
    }
    1;
}

sub _drop_from_index {
    my($idx, $obj_id, $col_val) = @_;
    $col_val = '' unless defined $col_val;
    return unless exists $idx->{$col_val};
    my $idx_val = $idx->{$col_val};
    $idx_val = '' unless defined $idx_val;
    my %ids = map { $_ => 1 } split /$;/, $idx_val;
    delete $ids{$obj_id};
    if (%ids) {
        $idx->{$col_val} = join $;, keys %ids;
    } else {
        delete $idx->{$col_val};
    }
}

sub generate_id {
    my $driver = shift;
    my($this) = @_;
    my $class = ref($this) || $this;
    my $id_file = File::Spec->catfile(
        $driver->cfg->DataSource, "ids.db");
    my($DB, $db) = _tie_db_file($id_file, $DB_HASH) or
        return $driver->error("Tie '$id_file' failed: $!");
    $db->{$class} = 0 unless exists $db->{$class};
    my $id = ++$db->{$class};
    undef $DB;
    untie %$db;
    $id;
}

1;
